import sys
import spotipy
from spotipy.oauth2 import SpotifyOAuth, CacheFileHandler
from tabulate import tabulate
from typing import List, Dict

# Local imports
from spotify_sync.config import Config
from spotify_sync.io_ import PersistentDataService
from spotify_sync.dataclasses import SpotifySong


class SpotifyService:
    def __init__(self, app_config: Config, pd_svc: PersistentDataService):
        from spotify_sync.log import get_logger

        self._logger = get_logger().getChild("SpotifyService")
        self.config = app_config
        self.pd_svc = pd_svc
        self.sp = spotipy.Spotify(auth_manager=self._get_oauth())
        self.paginator = SpotifyPaginator(self)
        self.songs = []

    def sync(self):
        self._preflight()
        cached = self.pd_svc.load_spotify_songs()
        online = self._fetch_liked()

        liked = self._merge_liked(cached, online)

        if self.config.data["SPOTIFY_PLAYLISTS_ENABLED"]:
            liked, playlists = self._merge_playlist_songs(liked)
            mapping = self._generate_playlist_songs_mapping(playlists)
            self.pd_svc.persist_playlist_mapping(mapping)

        self.pd_svc.persist_spotify_songs(liked)

    def cache_spotify_auth(self):
        auth = self._get_oauth()

        def _auth_headers():
            try:
                token = auth.get_access_token(as_dict=False)
            except TypeError:
                token = auth.get_access_token()
            return {"Authorization": "Bearer {0}".format(token)}

        headers = _auth_headers()
        return

    def display_playlist_stats(self):
        self._preflight()
        stats = self._get_playlist_stats()
        header = stats[0].keys()
        rows = [x.values() for x in stats]
        print()
        print(tabulate(rows, header))
        print()

    def _get_playlist_stats(self) -> List[dict]:
        ret = []

        playlists = self._get_all_playlists()
        for k in playlists:
            playlist = playlists[k]
            ret.append(
                {
                    "name": playlist["name"],
                    "id": playlist["id"],
                    "owner": self.paginator.is_playlist_owner(playlist),
                    "collaborative": playlist["collaborative"],
                    "public": playlist["public"],
                    "tracks": playlist["tracks"]["total"],
                    "url": playlist["external_urls"]["spotify"],
                }
            )

        return ret

    def _generate_playlist_songs_mapping(self, playlists: dict):
        ret = {}
        for k, playlist in playlists.items():
            ret[k] = {"playlist_name": playlist["name"], "playlist_id": k, "tracks": []}
            songs = self.paginator.get_songs_from_playlists(playlist=playlist)
            ret[k]["tracks"] = [v.id_ for v in songs]

        return ret

    def _get_all_playlists(self, playlists_to_exclude=[]):
        return self.paginator.user_playlists(playlists_to_exclude)

    def _get_all_playlist_songs(self, playlists):
        return self.paginator.get_songs_from_playlists(playlists=playlists)

    def _fetch_liked(self) -> List[SpotifySong]:
        self._logger.info("Fetching liked songs")
        return self.paginator.user_liked()

    def _merge_liked(self, cached: Dict[str, SpotifySong], online: List[SpotifySong]):
        ret = cached.copy()
        new_liked_songs = 0
        for s in online:
            if s.id_ not in ret:
                ret[s.id_] = s
                new_liked_songs += 1

        self._logger.info(
            f"Added {str(new_liked_songs)} new liked song(s) from Spotify (cached:{len(cached)})"
        )
        return ret

    def _merge_playlist_songs(self, songs: Dict[str, SpotifySong]):
        self._logger.info("Fetching playlist songs")
        playlists = self._get_all_playlists(
            self.config.data["SPOTIFY_PLAYLISTS_EXCLUDED"]
        )
        playlist_songs = self._get_all_playlist_songs(playlists)
        self._logger.debug(f"{len(playlist_songs)} playlist songs")
        added_playlist_songs = 0
        for s in playlist_songs:
            if s.id_ not in songs:
                songs[s.id_] = s
                added_playlist_songs += 1

        self._logger.info(
            f"Adding {added_playlist_songs}/{len(playlist_songs)} from playlists"
        )
        return songs, playlists

    def _get_oauth(self) -> SpotifyOAuth:
        handler = CacheFileHandler(
            cache_path=self.pd_svc.get_spotify_oauth(),
            username=self.config.data["SPOTIFY_USERNAME"],
        )

        return SpotifyOAuth(
            open_browser=False,
            # username=self.config.data["SPOTIFY_USERNAME"],
            scope=self.config.data["SPOTIFY_SCOPE"],
            client_id=self.config.data["SPOTIFY_CLIENT_ID"],
            client_secret=self.config.data["SPOTIFY_CLIENT_SECRET"],
            redirect_uri=f'http://127.0.0.1:{self.config.data["SPOTIFY_REDIRECT_URI_PORT"]}',
            # cache_path=self.pd_svc.get_spotify_oauth(),
            cache_handler=handler,
        )

    def _oauth_is_cached(self):
        return (
            self.pd_svc.get_spotify_oauth().exists()
            and self.pd_svc.get_spotify_oauth().is_file()
        )

    def _preflight(self):
        if not self._oauth_is_cached():
            command_sffx = (
                f"--profile {self.config.profile}"
                if self.config.profile is not None
                else f"--config {self.config.path}"
            )
            self._logger.error(
                f'No Spotify OAuth token found, please cache one using "spotify_sync utils authorize-spotify {command_sffx}"'
            )
            sys.exit(1)


class SpotifyPaginator:
    def __init__(self, spotify_svc: SpotifyService):
        from spotify_sync.log import get_logger

        self._logger = get_logger().getChild("SpotifyPaginator")
        self.spotify_svc = spotify_svc

    def user_liked(self) -> List[SpotifySong]:
        response = self._get_all("liked")
        return self._to_spotify_song(response)

    def user_playlists(self, playlists_to_exclude) -> dict:
        return self._get_all("playlists", playlists_to_exclude=playlists_to_exclude)

    def get_songs_from_playlists(
        self, playlist=None, playlists=None
    ) -> List[SpotifySong]:
        if playlist is not None:
            response = self._get_all("playlist-songs", playlist=playlist)
        else:
            response = self._get_all("playlist-songs", playlists=playlists)

        return self._to_spotify_song(response)

    def _exclude_playlists(self, playlists, playlists_to_exclude) -> dict:
        ret = {}
        for playlist in playlists:
            if playlist["name"] not in playlists_to_exclude:
                ret[playlist["id"]] = playlist
            else:
                self._logger.debug(f'{playlist["name"]} was in exclude list, skipping')

        return ret

    def _get_all(self, route: str, **kwargs) -> dict:
        assert route in ["liked", "playlists", "playlist-songs"]

        if route == "liked":
            return self._page_api(
                self.spotify_svc.sp.current_user_saved_tracks(limit=50)
            )

        if route == "playlists":
            playlists = self._page_api(
                self.spotify_svc.sp.current_user_playlists(limit=50)
            )
            if "playlists_to_exclude" in kwargs:
                playlists = self._exclude_playlists(
                    playlists, kwargs["playlists_to_exclude"]
                )

            if self.spotify_svc.config.data["SPOTIFY_PLAYLISTS_OWNER_ONLY"]:
                playlists = self._filter_playlists_by_owner(playlists)

            return playlists

        if route == "playlist-songs":
            if "playlist" not in kwargs and "playlists" not in kwargs:
                raise Exception(
                    "Must provide either playlist or playlists to extract songs from"
                )

            if "playlist" in kwargs:
                return self._page_api(
                    self.spotify_svc.sp.playlist_items(kwargs["playlist"]["id"])
                )

            ret = []
            for k, playlist in kwargs["playlists"].items():
                self._logger.debug(f'Fetching songs for playlist: {playlist["name"]}')
                ret.extend(
                    self._page_api(self.spotify_svc.sp.playlist_items(playlist["id"]))
                )

            return ret

    def _page_api(self, api_response: dict) -> dict:
        ret = api_response["items"]
        idx = 1
        while api_response["next"]:
            self._logger.debug(f"Sending Spotify API call {idx}")
            api_response = self.spotify_svc.sp.next(api_response)
            ret.extend(api_response["items"])
            idx += 1

        return ret

    @staticmethod
    def _to_spotify_song(api_songs: dict) -> List[SpotifySong]:
        ret = []
        for api_s in api_songs:
            s = SpotifySong()
            s.from_api(api_s)
            if s.is_valid():
                ret.append(s)

        return ret

    def is_playlist_owner(self, playlist: dict):
        user_id = self.spotify_svc.sp.current_user()["id"]
        return playlist["owner"]["id"] == user_id

    def _filter_playlists_by_owner(self, playlists):
        ret = {}
        for k, playlist in playlists.items():
            if self.is_playlist_owner(playlist):
                ret[k] = playlist
            else:
                self._logger.debug(
                    f'{playlist["name"]} is not owned by current user, skipping'
                )

        return ret



import json
import subprocess
import sys
from typing import Any, Dict
import hashlib
import tempfile
from pathlib import Path
import uuid

# local imports
from spotify_sync.dataclasses import SpotifySong, ProcessedSong


def as_spotify_song(songs: dict) -> Dict[str, SpotifySong]:
    ret = {}
    for k in songs:
        s = SpotifySong()
        s.from_dict(songs[k])
        ret[k] = s

    return ret


def as_processed_song(songs: dict) -> Dict[str, ProcessedSong]:
    ret = {}
    for k in songs:
        ret[k] = ProcessedSong(**songs[k])

    return ret


def get_md5(file):
    md5_hash = hashlib.md5()
    with open(file, "rb") as f:
        # Read and update hash in chunks of 4K
        for byte_block in iter(lambda: f.read(4096), b""):
            md5_hash.update(byte_block)
        return md5_hash.hexdigest()


def get_temp_dir() -> Path:
    dir_ = Path(tempfile.gettempdir())
    dir_.mkdir(exist_ok=True)
    return dir_


def get_temp_fn(extension=".txt") -> str:
    return str(uuid.uuid4()) + extension


def get_temp_file(extension=".txt") -> Path:
    temp_dir = get_temp_dir()
    return temp_dir / (str(uuid.uuid4()) + extension)


def load_json(file: Path) -> dict:
    if file.is_file():
        with open(file, mode="r", encoding="utf-8") as f:
            ret = json.load(f)
    else:
        ret = {}

    return ret


def dump_json(file: Path, data: Any, hidden=False, sort_keys=True) -> None:
    def obj_dict(obj):
        return obj.__dict__

    if hidden and file.exists():
        set_hidden_attribute(file, mode="unhide")

    with open(file, mode="w", encoding="utf-8") as f:
        json.dump(data, f, indent=4, sort_keys=sort_keys, default=obj_dict)

    if hidden:
        set_hidden_attribute(file, mode="hide")


def set_hidden_attribute(file: Path, mode: str):
    assert mode in ["hide", "unhide"]
    if sys.platform.startswith("win32") and mode == "hide":
        subprocess.check_call(["attrib", "+H", file])
    elif sys.platform.startswith("win32") and mode == "unhide":
        subprocess.check_call(["attrib", "-H", file])



import typing
import click

# local imports
from spotify_sync.cli import SpotifySyncApp
from spotify_sync.click_helper import (
    restore_zip_file,
    log_level,
    config_option_group,
    config_option_file,
    config_option_profile,
    spotify_songs_path,
    processed_songs_path,
    migrated_profile_name,
    migrated_config_schema_mode,
    restore_zip_file_force_overwrite,
    restore_zip_file_profile,
    backup_output_dir,
)

app = SpotifySyncApp()


@click.group()
def cli() -> None:
    """
    spotify_sync

    A schedulable, configurable CLI downloader for Spotify accounts
    """


@cli.group()
def run() -> None:
    """
    Run spotify_sync in different modes
    """


@run.command("auto")
@log_level
@config_option_group
@config_option_file
@config_option_profile
def auto(config, profile, verbose) -> None:
    """Runs spotify_sync in automatic mode"""
    app.auto()


@run.command("sync-spotify")
@config_option_group
@config_option_file
@config_option_profile
@log_level
def sync_spotify(config, profile, verbose) -> None:
    """Download Spotify song metadata associated with account"""
    app.sync_spotify()


@run.command("match-spotify")
@config_option_group
@config_option_file
@config_option_profile
@log_level
def match_spotify(config, profile, verbose) -> None:
    """Match unprocessed Spotify songs to Deezer equivalents"""
    app.match_spotify()


@run.command("download-missing")
@config_option_group
@config_option_file
@config_option_profile
@log_level
def download_missing(config, profile, verbose) -> None:
    """Downloads matched songs pending download"""
    app.download_missing()


@cli.group()
def utils() -> None:
    """
    Helper commands
    """


@utils.command("authorize-spotify")
@config_option_group
@config_option_file
@config_option_profile
@log_level
def authorize_spotify(config, profile, verbose) -> None:
    """Caches Spotify OAuth token"""
    app.authorize_spotify()


@utils.command("migrate-config")
@config_option_group
@config_option_file
def migrate_config(config) -> None:
    """Migrate legacy config file"""
    app.migrate_config()


@utils.command("migrate-to-profile")
@config_option_group
@config_option_file
@spotify_songs_path
@processed_songs_path
@migrated_profile_name
@migrated_config_schema_mode
def migrate_to_profile(config, spotify, processed, profile, config_schema_mode) -> None:
    """Migrate historical config and cache to a profile"""
    legacy = config_schema_mode == "legacy"
    app.migrate_to_profile(spotify, processed, profile, legacy)


@utils.command("manual-scan")
@click.argument("paths", required=True, type=click.STRING, nargs=-1)
def manual_plex_scan(paths: typing.Tuple[str]) -> None:
    """
    Manually invokes Autoscan with provided paths

    E.g. manual-scan /path/to/file/1 /path/to/file/2
    """
    app.scan([p for p in paths])


# @utils.command("validate-downloaded-files")
# @config_option_group
# @config_option_file
# @config_option_profile
# def validate_downloaded_files(config, profile) -> None:
#     """Validates processed songs are still on disk"""
#     app.validate_downloaded_files()


@cli.group()
def backup() -> None:
    """
    Backup config and persistent data
    """


@backup.command("local")
@config_option_group
@config_option_file
@config_option_profile
@backup_output_dir
def backup_local(config, profile, out_dir) -> None:
    """
    Backup to the current working directory
    """
    app.local_backup(out_dir)


@cli.command("restore")
@restore_zip_file
@restore_zip_file_profile
@restore_zip_file_force_overwrite
def restore(zip_file, new_profile, force) -> None:
    """
    Restore an existing backup
    """
    app.local_restore(zip_file, force=force, new_profile=new_profile)


@cli.group()
def stats() -> None:
    """
    Display stats from various sources
    """


@stats.command("playlists")
@config_option_group
@config_option_file
@config_option_profile
def playlist_stats(config, profile) -> None:
    """
    Display playlist stats associated with account
    """
    app.playlist_stats()


@stats.command("failed-download")
@config_option_group
@config_option_file
@config_option_profile
def failed_download_stats(config, profile) -> None:
    """
    Display historical failed song download stats
    """
    app.failed_download_stats()


@cli.group()
def config() -> None:
    """
    Access app configurations
    """


@config.command("add")
@click.argument("name", required=True)
@click.argument("path", required=True, type=click.Path(exists=True))
def config_add(name, path) -> None:
    """
    Add a config profile to the cache
    """
    app.cache_config_profile(name, path)


@config.command("remove")
@click.argument("name", required=True)
def config_remove(name) -> None:
    """
    Remove a config profile from the cache
    """
    app.remove_config_profile(name)


@config.command("list")
def config_list() -> None:
    """
    List cached config profiles
    """
    app.list_config_profiles()


@config.command("list-paths")
def config_list_paths() -> None:
    """
    List cached config profile paths
    """
    app.list_config_profile_paths()


@config.command("generate")
def config_generate() -> None:
    """
    Generate an example config file
    """
    app.generate_example_config()


if __name__ == "__main__":
    cli()



import sys
import os
from typing import List, Dict, Tuple
from concurrent.futures import ThreadPoolExecutor, as_completed, wait

# deemix
from deezer import Deezer
from deemix import generateDownloadObject
from deemix.downloader import Downloader
from deemix.utils import getBitrateNumberFromText
from deemix.types.DownloadObjects import Single

# local imports
from spotify_sync.config import Config
from spotify_sync.dataclasses import ProcessedSong, DownloadStatus
from spotify_sync.common import get_md5, dump_json, load_json

arl_valid = False


# bitrate_name_to_number = {
#     '360': 15,
#     '360_mq': 14,
#     '360_lq': 13,
#     'lossless': 9,
#     '320': 3,
#     '128': 1
# }


class LogListener:
    def __init__(self):
        self.messages = []

    def send(self, key, value=None):
        self.messages.append({key: value})


class DownloadLogger:
    def __init__(self, index: int, total: int, song: ProcessedSong, logger):
        self._logger = logger
        self._downloadIndex = index
        self._downloadTotal = total
        self._song = song
        self._base_message: str = ""
        self._get_base_message()

    def _get_base_message(self):
        if len(str(self._downloadIndex)) != len(str(self._downloadTotal)):
            zero_to_pad = len(str(self._downloadTotal)) - len(str(self._downloadIndex))
            self._base_message = f'{" ": <{zero_to_pad}}{str(self._downloadIndex)}/{str(self._downloadTotal)}'
        else:
            self._base_message = (
                f"{str(self._downloadIndex)}/{str(self._downloadTotal)}"
            )

    @staticmethod
    def _get_message(message):
        return f"- {message}" if message else ""

    def info(self, action: str, message: str):
        self._logger.info(
            f"{self._base_message} - {action} [{self._song.spotify_artist} - {self._song.spotify_title}] {self._get_message(message)}"
        )

    def warn(self, action: str, message: str):
        self._logger.warning(
            f"{self._base_message} - {action} [{self._song.spotify_artist} - {self._song.spotify_title}] {self._get_message(message)}"
        )

    def error(self, action: str, message: str):
        self._logger.error(
            f"{self._base_message} - {action} [{self._song.spotify_artist} - {self._song.spotify_title}] {self._get_message(message)}"
        )


class DeemixHelper:
    def __init__(self, app_config: Config):
        from spotify_sync.log import get_logger

        self._logger = get_logger().getChild("DeemixHelper")
        self.config = app_config

    def arl_valid(self) -> bool:
        self._logger.debug(f"Checking if arl is valid")
        global arl_valid
        if not arl_valid:
            self._logger.debug(f"arl_valid is False")
            arl = self.config.data["DEEMIX_ARL"].strip()
            self._logger.debug(f"Logging in with arl in config.json")
            client = Deezer()
            login = client.login_via_arl(arl)

            if login:
                self._logger.debug(f"Login successful")
                arl_valid = True
                return True
            else:
                self._logger.error(
                    f"Failed to login to Deezer with arl, you may need to refresh it"
                )
                return False

    def check_deemix_config(self):
        if os.path.isfile(self.config.data["DEEMIX_DOWNLOAD_PATH"]):
            self._logger.error(
                f'{self.config.data["DEEMIX_DOWNLOAD_PATH"]} is a file, must not exist or be an existent folder'
            )
            sys.exit(1)

        if not os.path.isdir(self.config.data["DEEMIX_DOWNLOAD_PATH"]):
            self._logger.warning(
                f'{self.config.data["DEEMIX_DOWNLOAD_PATH"]} does not exist, creating'
            )
            os.mkdir(self.config.data["DEEMIX_DOWNLOAD_PATH"])

        if "\\" in self.config.data["DEEMIX_DOWNLOAD_PATH"]:
            self.config.data["DEEMIX_DOWNLOAD_PATH"] = self.config.data[
                "DEEMIX_DOWNLOAD_PATH"
            ].replace("\\", "/")

        accepted_bitrates = ["lossless", "320", "360", "360_mq", "360_lq", "128"]
        if self.config.data["DEEMIX_MAX_BITRATE"] not in accepted_bitrates:
            self._logger.error(
                f'{self.config.data["DEEMIX_MAX_BITRATE"]} must be one of {",".join(accepted_bitrates)}'
            )
            sys.exit(1)

    def get_deemix_config(
        self,
    ) -> dict:
        deemix_config = DEFAULTS.copy()
        deemix_config["downloadLocation"] = self.config.data["DEEMIX_DOWNLOAD_PATH"]
        deemix_config["maxBitrate"] = getBitrateNumberFromText(
            self.config.data["DEEMIX_MAX_BITRATE"]
        )
        deemix_config["fallbackISRC"] = not self.config.data["DEEMIX_STRICT_MATCHING"]
        return deemix_config


class DeemixDownloader:
    def __init__(self, app_config: Config):
        from spotify_sync.log import get_logger

        self._logger = get_logger().getChild("DeemixDownloader")
        self.dz = Deezer()
        self.app_config = app_config
        self.config = DeemixHelper(self.app_config).get_deemix_config()
        self.deezer_logged_in = self.dz.login_via_arl(
            self.app_config.data["DEEMIX_ARL"]
        )
        self.skip_low_quality = self.app_config.data["DEEMIX_SKIP_LOW_QUALITY"]
        self.max_bitrate = self.app_config.data["DEEMIX_MAX_BITRATE"]
        self.songs_to_download: List[ProcessedSong] = []
        self.download_report: Dict[str, DownloadStatus] = {}

    def download_wrapper(self, index, num_urls_to_download, song, download_obj):
        logger = DownloadLogger(
            index=index, total=num_urls_to_download, song=song, logger=self._logger
        )
        listener = LogListener()
        dl = Downloader(self.dz, download_obj, self.config, listener)

        logger.info(action="STARTING", message="")
        dl.start()

        self.update_download_report(dl.downloadObject, song, listener, logger)

    def download_songs(self, songs: List[ProcessedSong]):
        self.songs_to_download = songs

        if not self.deezer_logged_in:
            self._logger.error("Failed to login with arl, you may need to refresh it")
            sys.exit(1)

        if self.skip_low_quality and self.max_bitrate in ["lossless", "320"]:
            if self.max_bitrate == "lossless" and not self.dz.current_user.get(
                "can_stream_lossless"
            ):
                self._logger.info(
                    f"SKIP_LOW_QUALITY is specified and unable to stream FLAC, stopping script! "
                    f"If this is unexpected, please ensure your Deezer account is Premium/Hi-Fi"
                )
                sys.exit(1)

            if self.max_bitrate == "320" and not self.dz.current_user.get(
                "can_stream_hq"
            ):
                self._logger.info(
                    f"SKIP_LOW_QUALITY is specified and unable to stream 320, stopping script1 "
                    f"If this is unexpected, please ensure your Deezer account is Premium/Hi-Fi"
                )
                sys.exit(1)

        self._logger.info(f"Gathering song information in preparation for download..")
        download_objs = []
        with ThreadPoolExecutor(self.app_config.data["THREADS"]) as executor:
            dl_obj_futures = {
                executor.submit(
                    generateDownloadObject,
                    self.dz,
                    v.deezer_url,
                    self.config["maxBitrate"],
                ): v
                for v in self.songs_to_download
            }
            for future in as_completed(dl_obj_futures):
                song = dl_obj_futures[future]
                download_obj = future.result()
                download_objs.append({"song": song, "download_obj": download_obj})

        num_urls_to_download = len(self.songs_to_download)
        with ThreadPoolExecutor(self.app_config.data["THREADS"]) as executor:
            dl_futures = [
                executor.submit(
                    self.download_wrapper,
                    i + 1,
                    num_urls_to_download,
                    v["song"],
                    v["download_obj"],
                )
                for i, v in enumerate(download_objs)
            ]
            wait(dl_futures)

    @staticmethod
    def download_skipped(listener: LogListener):
        download_skipped = False
        for m in listener.messages:
            if "downloadInfo" in m:
                if m["downloadInfo"].get("state") == "alreadyDownloaded":
                    download_skipped = True

        return download_skipped

    def update_download_report(
        self,
        download_object,
        requested_song: ProcessedSong,
        listener: LogListener,
        dl_logger,
    ):
        if not isinstance(download_object, Single):
            self._logger.error("Not a Single, unexpected type!", type(download_object))
            sys.exit(1)

        errors = None
        md5 = ""
        status = False
        f = ""
        downloaded_isrc = None
        downloaded_link = None
        downloaded_id = None
        download_skipped = self.download_skipped(listener)

        if download_object.downloaded == 1:
            downloaded_isrc = download_object.single["trackAPI"]["isrc"]
            downloaded_link = download_object.single["trackAPI"]["link"]
            downloaded_id = download_object.single["trackAPI"]["id"]
            f = download_object.files[0]["path"]

            if download_skipped:
                dl_logger.info(
                    action="FINSIHED", message=f"Skipping, already downloaded"
                )
                status = True
                md5 = get_md5(f)

            elif os.path.isfile(f):
                dl_logger.info(action="FINSIHED", message=f"Successfully downloaded")
                status = True
                md5 = get_md5(f)
            else:
                dl_logger.warn(
                    action="FINSIHED",
                    message=f"Failed, downloaded but could not find {f}",
                )
                errors = [f"Downloaded but could not find {f}"]

        elif download_object.failed == 1:
            dl_logger.warn(
                action="FINSIHED",
                message=f'Failed, download for DeezerId: {download_object.single["trackAPI"]["id"]} error: {download_object.errors[0]["message"]}',
            )
            errors = download_object.errors

        self.download_report[requested_song.spotify_id] = DownloadStatus(
            spotify_id=requested_song.spotify_id,
            deezer_id=downloaded_id,
            requested_isrc=requested_song.deezer_isrc,
            downloaded_isrc=downloaded_isrc,
            requested_url=requested_song.deezer_url,
            downloaded_url=downloaded_link,
            requested_bitrate=self.config["maxBitrate"],
            downloaded_bitrate=download_object.bitrate,
            success=status,
            skipped=download_skipped,
            errors=errors or [],
            download_path=f,
            md5=md5,
        )

    def get_report(self) -> Tuple[int, int, List[DownloadStatus]]:
        count_of_failed = 0
        count_of_succeeded = 0
        reports = []

        for v in self.download_report.values():
            if not v.success:
                count_of_failed += 1
            else:
                count_of_succeeded += 1

            reports.append(v)

        return count_of_failed, count_of_succeeded, reports

    @staticmethod
    def extract_isrc_from_download_object(obj):
        return obj.single["trackAPI"]["isrc"]


DEFAULTS = {
    "downloadLocation": "DOWNLOAD_LOCATION_PATH",
    "tracknameTemplate": "%artist% - %title%",
    "albumTracknameTemplate": "%artist% - %title%",
    "playlistTracknameTemplate": "%position% - %artist% - %title%",
    "createPlaylistFolder": False,
    "playlistNameTemplate": "%playlist%",
    "createArtistFolder": True,
    "artistNameTemplate": "%artist%",
    "createAlbumFolder": True,
    "albumNameTemplate": "%album%",
    "createCDFolder": False,
    "createStructurePlaylist": True,
    "createSingleFolder": True,
    "padTracks": True,
    "paddingSize": "0",
    "illegalCharacterReplacer": "_",
    "maxBitrate": "9",  # default to lossless
    "feelingLucky": False,
    "fallbackBitrate": True,  # Allow lower quality than target
    "fallbackSearch": False,
    "fallbackISRC": True,
    "featuredToTitle": True,
    "logErrors": True,
    "logSearched": False,
    "overwriteFile": False,
    "createM3U8File": False,
    "playlistFilenameTemplate": "playlist",
    "syncedLyrics": False,
    "embeddedArtworkSize": 800,
    "embeddedArtworkPNG": False,
    "localArtworkSize": 1400,
    "localArtworkFormat": "jpg",
    "saveArtwork": False,
    "coverImageTemplate": "cover",
    "saveArtworkArtist": False,
    "artistImageTemplate": "folder",
    "jpegImageQuality": 90,
    "dateFormat": "Y-M-D",
    "albumVariousArtists": True,
    "removeAlbumVersion": False,
    "removeDuplicateArtists": True,
    "titleCasing": "nothing",
    "artistCasing": "nothing",
    "executeCommand": "",
    "tags": {
        "title": True,
        "artist": True,
        "artists": True,
        "album": True,
        "cover": True,
        "trackNumber": True,
        "trackTotal": False,
        "discNumber": True,
        "discTotal": False,
        "albumArtist": True,
        "genre": True,
        "year": True,
        "date": True,
        "explicit": False,
        "isrc": True,
        "length": True,
        "barcode": False,
        "bpm": True,
        "replayGain": False,
        "label": True,
        "lyrics": False,
        "syncedLyrics": False,
        "copyright": False,
        "composer": False,
        "involvedPeople": False,
        "source": False,
        "rating": False,
        "savePlaylistAsCompilation": False,
        "useNullSeparator": False,
        "saveID3v1": True,
        "multiArtistSeparator": "default",
        "singleAlbumArtist": False,
        "coverDescriptionUTF8": False,
    },
}



import errno
import logging
import sys
import os
from pathlib import Path
from logging.handlers import RotatingFileHandler

ROOT_LOGGER = None
LOG_LEVEL = (
    os.environ["LOG_LEVEL"] if os.environ.get("LOG_LEVEL") is not None else "INFO"
)
LOG_PATH = (
    Path(os.environ["LOG_PATH"]) if os.environ.get("LOG_PATH") is not None else ""
)


def get_logger() -> logging.Logger:
    global ROOT_LOGGER

    if ROOT_LOGGER is None:
        # Configure root logger
        log_formatter = logging.Formatter(
            "[%(asctime)s] %(levelname)-9s %(name)-16s : %(message)s"
        )
        ROOT_LOGGER = logging.getLogger()
        ROOT_LOGGER.setLevel(LOG_LEVEL)

        # Console logger, log to stdout instead of stderr
        consoleHandler = logging.StreamHandler(sys.stdout)
        consoleHandler.setFormatter(log_formatter)
        ROOT_LOGGER.addHandler(consoleHandler)

        # File logger
        if LOG_PATH == "":
            log_file = Path(os.getcwd()) / "spotify_sync.log"
        else:
            log_file = LOG_PATH

        if not os.path.exists(os.path.dirname(log_file)):
            try:
                os.makedirs(os.path.dirname(log_file))
            except OSError as exc:  # Guard against race condition
                if exc.errno != errno.EEXIST:
                    raise

        fileHandler = RotatingFileHandler(
            log_file, maxBytes=2097152, backupCount=5, encoding="utf-8"
        )

        fileHandler.setFormatter(log_formatter)
        ROOT_LOGGER.addHandler(fileHandler)

    return ROOT_LOGGER


def _setup_external_logging() -> None:
    # Set external module logging policy
    if LOG_LEVEL == "DEBUG":
        # logging.getLogger('urllib3').setLevel(logging.DEBUG)
        logging.getLogger("urllib3").setLevel(logging.ERROR)
        logging.getLogger("spotipy").setLevel(logging.DEBUG)
        logging.getLogger("deemix").propagate = True
    else:
        # Decrease modules logging
        logging.getLogger("urllib3").setLevel(logging.ERROR)
        logging.getLogger("spotipy").setLevel(logging.ERROR)
        logging.getLogger("deemix").propagate = False


_setup_external_logging()



from typing import Dict
import deezer as Deezer
from deezer.errors import DataException
from concurrent.futures import ThreadPoolExecutor, wait

# local imports
from spotify_sync.dataclasses import Config
from spotify_sync.io_ import PersistentDataService
from spotify_sync.dataclasses import ProcessedSong, SpotifySong

# load Deezer (for matches), API is unauthenticated
client = Deezer.Deezer()


class MatchLogger:
    def __init__(self, song: SpotifySong):
        from spotify_sync.log import get_logger

        self._logger = get_logger().getChild("Match")
        self._song = song

    def info(self, message: str):
        self._logger.info(f"[{self._song.artist} - {self._song.title}] - {message}")

    def warning(self, message: str):
        self._logger.warning(f"[{self._song.artist} - {self._song.title}] - {message}")

    def debug(self, message: str):
        self._logger.debug(f"[{self._song.artist} - {self._song.title}] - {message}")

    def error(self, message: str):
        self._logger.error(f"[{self._song.artist} - {self._song.title}] - {message}")


class SongMatcher:
    def __init__(self, song: SpotifySong):
        self._logger = MatchLogger(song)
        self.song = song
        self.match = False
        self.match_type = None
        self.match_message = None
        self.match_payload = None

    def _match_via_isrc(self):
        track = Deezer.API.get_track(client.api, f"isrc:{self.song.isrc}")

        self._logger.debug(f"Matched - [isrc]")
        self.match = True
        self.match_type = "isrc"
        self.match_payload = track

    def _match_fuzzy(self):
        deezer_search = Deezer.API.advanced_search(
            client.api,  # self
            self.song.artist,  # artist
            self.song.album,  # album
            self.song.title,  # track name
        )

        if len(deezer_search["data"]) > 0:
            self._logger.debug(f"Matched - [fuzzy]")
            self.match = True
            self.match_type = "fuzzy"
            self.match_payload = deezer_search["data"][0]
        else:
            self._logger.debug(f"Failed fuzzy searching, SpotifyId: {self.song.id_}")

    def search(self):
        try:
            self._match_via_isrc()
        except DataException:
            self._match_fuzzy()
        except Exception as ex:
            raise Exception(ex)

        if self.match:
            if "link" not in self.match_payload:
                self._logger.warning(
                    f"[SpotifyId:{self.song.id_}] - Matched but response does not contain a Deezer link, unmatching.."
                )
                self.match = False
                self.match_message = (
                    "Matched but response does not contain a Deezer link"
                )
                return

            if self.match_payload.get("artist") is not None and not self.match_payload[
                "artist"
            ].get("name"):
                self._logger.warning(
                    f"[SpotifyId:{self.song.id_}] - Matched but response does not contain an Deezer artist, unmatching.."
                )
                self.match = False
                self.match_message = f"Matched but response does not contain a artist name, likely this link will not work. Matched link: '{self.match_payload['link']}'"
                return

            if "id" not in self.match_payload:
                self._logger.warning(
                    f"[SpotifyId:{self.song.id_}] - Matched but response does not contain a Deezer id, unmatching.."
                )
                self.match = False
                self.match_message = (
                    f"Matched but response does not contain a Deezer id"
                )


class MatchService:
    def __init__(self, app_config: Config, pd_svc: PersistentDataService):
        from spotify_sync.log import get_logger

        self._logger = get_logger().getChild("MatchService")
        self.config = app_config
        self.pd_svc = pd_svc
        self.spotify: Dict[str, SpotifySong] = {}
        self.processed: Dict[str, ProcessedSong] = {}
        self.unprocessed: Dict[str, SpotifySong] = {}
        self.matched_count = 0

    def process_spotify(self):
        self.pd_svc.verify_files()
        self._refresh_files()
        self._match_unprocessed()
        self.pd_svc.persist_processed_songs(self.processed)

    def _refresh_files(self):
        self.spotify = self.pd_svc.load_spotify_songs()
        self.processed = self.pd_svc.load_processed_songs()
        self.unprocessed = self._get_unprocessed_songs()

    def _get_unprocessed_songs(self) -> Dict[str, SpotifySong]:
        self.unprocessed = {}
        for k in self.spotify:
            if k not in self.processed:
                self.unprocessed[k] = self.spotify[k]

        return self.unprocessed

    def _process_match(self, matcher: SongMatcher) -> bool:
        if matcher.match:
            if "link" not in matcher.match_payload:
                self._logger.warning(
                    f"[SpotifyId:{matcher.song.id_}] - Matched but response does not contain a Deezer link, unmatching.."
                )
                matcher.match = False
                matcher.match_message = (
                    "Matched but response does not contain a Deezer link"
                )

        s = ProcessedSong(
            spotify_title=matcher.song.title,
            spotify_artist=matcher.song.artist,
            spotify_isrc=matcher.song.isrc,
            spotify_url=matcher.song.url,
            spotify_id=matcher.song.id_,
            deezer_title=matcher.match_payload["title"] if matcher.match else None,
            deezer_artist=matcher.match_payload["artist"]["name"]
            if matcher.match
            else None,
            # deezer fuzzy search payload doesnt return isrc of matched item
            deezer_isrc=matcher.match_payload.get("isrc") if matcher.match else None,
            deezer_url=matcher.match_payload["link"] if matcher.match else None,
            deezer_id=matcher.match_payload["id"] if matcher.match else None,
            matched=matcher.match,
            match_type=matcher.match_type,
            match_message=matcher.match_message,
            match_pending_download=matcher.match,
        )

        self.processed[s.spotify_id] = s
        return s.matched

    def _match_unprocessed(self):
        self._logger.info(
            f"Attempting to match {len(self.unprocessed)} Spotify songs to Deezer"
        )

        matchers = [SongMatcher(song=song) for song in self.unprocessed.values()]
        with ThreadPoolExecutor(self.config.data["THREADS"]) as executor:
            match_futures = [executor.submit(m.search()) for m in matchers]
            wait(match_futures)

        for m in matchers:
            if self._process_match(m):
                self.matched_count += 1

        self._logger.info(
            f"Matched {self.matched_count}/{len(self.unprocessed)} new liked song(s)"
        )

    def _save_processed(self):
        self._logger.debug(
            f"Serializing {str(len(self.processed))} processed songs back to disk"
        )
        self.pd_svc.persist_processed_songs(self.processed)



from attr import dataclass
from typing import List, Dict, Union
from pathlib import Path


@dataclass
class Config:
    profile: Union[str, None]
    path: Path
    data: dict


@dataclass
class ConfigProfile:
    name: str
    path: Path


@dataclass
class DownloadStatus:
    spotify_id: str
    deezer_id: str
    requested_isrc: str
    downloaded_isrc: str
    requested_url: str
    downloaded_url: str
    requested_bitrate: str
    downloaded_bitrate: str
    success: bool
    skipped: bool
    errors: List[str]
    download_path: str
    md5: str


@dataclass
class BackupManifest:
    profile: Union[str, None]
    spotify_user: str
    timestamp: str
    checksum: str


@dataclass
class BackupSnapshot:
    profile: Union[str, None]
    username: str
    config: Path
    spotify: Path
    processed: Path
    playlists: Path
    oauth: Path


@dataclass
class ProcessedSong:
    """Class for keeping track of an offline song."""

    spotify_title: str = None
    spotify_artist: str = None
    spotify_isrc: str = None
    spotify_url: str = None
    spotify_id: str = None
    deezer_title: str = None
    deezer_artist: str = None
    deezer_isrc: str = None
    deezer_url: str = None
    deezer_id: str = None
    matched: bool = None
    match_type: str = None
    match_message: str = None
    match_pending_download: bool = None
    downloaded: str = None
    download_isrc: str = None
    download_url: str = None
    download_path: str = None
    download_md5: str = None
    download_bitrate: str = None
    download_failed: bool = None
    download_failed_reason: str = None


class SpotifySong:
    def __init__(
        self,
        id_=None,
        artist=None,
        album=None,
        title=None,
        url=None,
        isrc=None,
        explicit=None,
    ):
        self.id_: id_
        self.artist: artist
        self.album: album
        self.title: title
        self.url: url
        self.isrc: isrc
        self.explicit: explicit

    def from_api(self, song: dict):
        self.id_ = song["track"]["id"] if song.get("track") is not None else None
        self.artist = (
            song["track"]["artists"][0]["name"]
            if song.get("track") is not None
            else None
        )
        self.album = (
            song["track"]["album"]["name"] if song.get("track") is not None else None
        )
        self.title = song["track"]["name"] if song.get("track") is not None else None
        self.url = (
            song["track"]["external_urls"].get("spotify")
            if song.get("track") is not None
            else None
        )
        self.isrc = (
            song["track"]["external_ids"].get("isrc")
            if song.get("track") is not None
            else None
        )
        self.explicit = (
            song["track"]["explicit"] if song.get("track") is not None else None
        )

    def from_dict(self, dictionary: dict):
        self.id_ = dictionary["id_"]
        self.artist = dictionary["artist"]
        self.album = dictionary["album"]
        self.title = dictionary["title"]
        self.url = dictionary["url"]
        self.isrc = dictionary["isrc"]
        self.explicit = dictionary["explicit"]

    def is_valid(self):
        if self.id_ is not None and self.isrc is not None:
            return True

        return False



import requests
from pathlib import Path

# local imports
from spotify_sync.config import Config


class AutoScanService:
    def __init__(self, app_config: Config):
        from spotify_sync.log import get_logger

        self._logger = get_logger().getChild("AutoScanService")
        self.config = app_config

    def scan(self, paths):
        if self.config.data["AUTOSCAN_ENABLED"]:
            unique = set()
            for p in paths:
                if Path(p).is_file():
                    p = Path(p).parent
                unique.add(p)

            for p in unique:
                params = {"dir": p}

                if self.config.data["AUTOSCAN_AUTH_ENABLED"]:
                    response = requests.post(
                        self.config.data["AUTOSCAN_ENDPOINT"],
                        params=params,
                        auth=(
                            self.config.data["AUTOSCAN_USERNAME"],
                            self.config.data["AUTOSCAN_PASSWORD"],
                        ),
                    )
                else:
                    response = requests.post(
                        self.config.data["AUTOSCAN_ENDPOINT"], params=params
                    )

                if response.status_code == 200:
                    self._logger.debug(f"Plex scan request: {p}")
                else:
                    self._logger.error(f"Failed to send Plex scan notification: {p}")
            self._logger.info(f"Processed {len(paths)} scan request(s)")



import sys
from typing import List
from tabulate import tabulate

# local imports
from spotify_sync.dataclasses import ProcessedSong
from spotify_sync.deemix_ import DownloadStatus, DeemixDownloader
from spotify_sync.io_ import PersistentDataService
from spotify_sync.config import Config
from spotify_sync.deemix_ import DeemixHelper
from spotify_sync.pushover_ import PushoverClient


class DownloadService:
    def __init__(
        self,
        app_config: Config,
        pd_svc: PersistentDataService,
        pushover: PushoverClient,
    ):
        from spotify_sync.log import get_logger

        self._logger = get_logger().getChild("DownloadService")
        self._pushover = pushover
        self.config = app_config
        self.pd_svc = pd_svc
        self.downloaded_song_paths = []

    def preflight(self) -> None:
        helper = DeemixHelper(self.config)
        if not helper.arl_valid():
            self._pushover.send_message("Spotify downloader", "Failed to validate arl")
            sys.exit(1)

        helper.check_deemix_config()

    def get_songs_to_download(self) -> List[ProcessedSong]:
        processed = self.pd_svc.load_processed_songs()

        ret = []
        for k, song in processed.items():
            if song.match_pending_download and not song.download_failed:
                self._logger.debug(f"{k} is matched and awaiting download")
                ret.append(song)

        return ret

    def persist_download_status(self, download_statuses) -> None:
        processed = self.pd_svc.load_processed_songs()

        for status in download_statuses:
            if status.success:
                processed[status.spotify_id].match_pending_download = False
                processed[status.spotify_id].downloaded = True
                processed[status.spotify_id].download_isrc = status.downloaded_isrc
                processed[status.spotify_id].download_url = status.downloaded_url
                processed[status.spotify_id].download_path = status.download_path
                processed[status.spotify_id].download_md5 = status.md5
                processed[
                    status.spotify_id
                ].download_bitrate = status.downloaded_bitrate
                processed[status.spotify_id].download_failed = False
                processed[status.spotify_id].download_failed_reason = None
            else:
                processed[status.spotify_id].match_pending_download = True
                processed[status.spotify_id].downloaded = False
                processed[status.spotify_id].download_isrc = status.downloaded_isrc
                processed[status.spotify_id].download_url = status.downloaded_url
                processed[status.spotify_id].download_path = status.download_path
                processed[status.spotify_id].download_md5 = status.md5
                processed[status.spotify_id].download_failed = True
                processed[status.spotify_id].download_failed_reason = "\n".join(
                    [v["message"] for v in status.errors]
                )

        self.pd_svc.persist_processed_songs(processed)

    def get_failed_download_stats(self) -> List[dict]:
        ret = []
        for k, song in self.pd_svc.load_processed_songs().items():
            if song.download_failed:
                ret.append(
                    {
                        "spotifyId": song.spotify_id,
                        "deezerId": song.deezer_id,
                        "title": song.deezer_title,
                        "artist": song.deezer_artist,
                        "error": song.download_failed_reason,
                    }
                )

        return ret

    def get_failed_download_status_summary(self) -> List[dict]:
        ret = {}
        processed_songs = self.pd_svc.load_processed_songs()
        for k in processed_songs:
            song = processed_songs[k]
            if song.download_failed:
                if song.download_failed_reason not in ret:
                    ret[song.download_failed_reason] = 1
                else:
                    ret[song.download_failed_reason] += 1

        return [{"Failed downloads": v, "Reason": k} for k, v in ret.items()]

    def display_failed_download_stats(self) -> None:
        stats = self.get_failed_download_status_summary()
        if len(stats) == 0:
            print(f"No failed downloads")
            return

        header = stats[0].keys()
        rows = [x.values() for x in stats]
        print()
        print(tabulate(rows, header))
        print()

    def get_file_download_paths(self, download_report: List[DownloadStatus]):
        for status in download_report:
            if status.success:
                self.downloaded_song_paths.append(status.download_path)

    def download_missing_tracks(self) -> None:
        songs = self.get_songs_to_download()

        self._logger.info(f"{len(songs)} tracks pending download")
        if not songs:
            return

        self._logger.info(f"Downloading {len(songs)} song(s) from Deezer")
        downloader = DeemixDownloader(app_config=self.config)
        downloader.download_songs(songs)
        failed_songs, succeeded_songs, reports = downloader.get_report()
        self._logger.info(f"Successfully downloaded {succeeded_songs}/{len(songs)}")

        self.persist_download_status(reports)
        self.get_file_download_paths(reports)



from chump import Application

# local imports
from spotify_sync.config import Config


class PushoverClient:
    def __init__(self, app_config: Config = None):
        from spotify_sync.log import get_logger

        self.config = app_config
        self._logger = get_logger().getChild("PushoverClient")
        self._app = None
        self._user = None
        self.valid = False
        self.enabled = False
        self.initialize()

    def _validate(self):
        if not self._user.is_authenticated or not self._app.is_authenticated:
            self._logger.warning("Failed to validate Pushover connection details")
        else:
            self.valid = True

    def initialize(self):
        if self.config is not None and self.config.data["PUSHOVER_ENABLED"]:
            self.enabled = True
            self._app = Application(self.config.data["PUSHOVER_API_TOKEN"])
            self._user = self._app.get_user(self.config.data["PUSHOVER_USER_KEY"])
            self._validate()

    def send_message(self, title: str, message: str):
        if self.enabled:
            if self.valid:
                self._user.send_message(title=title, message=message)
            else:
                self._logger.warning(
                    f"Pushover details are invalid, unable to send: {message}"
                )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['spotify_sync', 'spotify_sync.backup']

package_data = \
{'': ['*'], 'spotify_sync': ['data/*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0',
 'chump>=1.6.0,<2.0.0',
 'click-option-group>=0.5.3,<0.6.0',
 'click>=8.1.3,<9.0.0',
 'deezer-py>=1.3.7,<2.0.0',
 'flatdict>=4.0.1,<5.0.0',
 'jsonschema>=4.4.0,<5.0.0',
 'psutil>=5.9.0,<6.0.0',
 'requests>=2.25.1,<3.0.0',
 'spotipy>=2.19.0,<3.0.0',
 'ss-deemx>=3.6.8,<4.0.0',
 'tabulate>=0.8.7,<0.9.0']

extras_require = \
{':sys_platform == "windows"': ['pywin32>=304,<305']}

entry_points = \
{'console_scripts': ['spotify_sync = spotify_sync.__main__:cli']}

setup_kwargs = {
    'name': 'spot-sync',
    'version': '1.0.0',
    'description': 'A schedulable, configurable CLI downloader for Spotify accounts',
    'long_description': '# spotify_sync\n[![](https://img.shields.io/pypi/pyversions/spotify_sync)](https://pypi.org/project/spotify-sync/)\n[![](https://img.shields.io/badge/docs-%20-yellow)](https://docs.spotify-sync.jbh.cloud/)\n[![](https://img.shields.io/github/license/jbh-cloud/spotify_sync)](https://github.com/jbh-cloud/spotify_sync/blob/main/LICENSE.md)\n![](https://img.shields.io/github/stars/jbh-cloud/spotify_sync?style=social)\n\n![](run_example.gif)\n\n## Introduction\n\nspotify_sync is a CLI app written in Python that allows you to download songs from your Spotify account. It is designed to be a \'set and forget\' tool for users wanting to keep an offline copy of their songs.\n\n#### Features:\n\n* Download of liked songs\n* Download of playlist songs\n* Up-to lossless quality downloads\n* Multi-threaded downloading\n* Scheduling (e.g. cron)\n* Multi-config support; configure and schedule multiple profiles with separate Spotify accounts\n* Backup and restore of config and persistent data\n* Notification support via [Pushover](https://pushover.net/)\n* Automatic Plex library scanning via [Autoscan](https://github.com/Cloudbox/autoscan)\n\n\n## Requirements\n1. Python & pip >= 3.8\n2. Spotify account (Free)\n3. Deezer account (Free allows 128kbps downloads, up to lossless requires Deezer Hi-Fi account)\n\n\n## Install\n\n```\npython3 -m pip install -U spotify_sync\n```\n\n## Usage\n\n*Simple usage would be..*\n\nCache Spotify OAuth token\n```\nspotify_sync utils authorize-spotify --profile myFirstProfile\n```\n\nRun in automatic mode\n```\nspotify_sync run auto --profile myFirstProfile\n```\n\n### Documentation\n\nFurther configuration is required, details for which can be found at the [docs](https://docs.spotify-sync.jbh.cloud/).\n\n\n<a href="https://www.buymeacoffee.com/jbhcloud" target="_blank"><img src="https://cdn.buymeacoffee.com/buttons/v2/default-yellow.png" alt="Buy Me A Coffee" style="height: 60px !important;width: 217px !important;" ></a>\n\n\n### Disclaimer\n\nThis tool was written for educational purposes. I will not be responsible if you use this program in bad faith. By using it, you are accepting the [Deezer Terms of Use](https://www.deezer.com/legal/cgu).\n    spotify_sync is not affiliated with Deezer.\n',
    'author': 'JBH',
    'author_email': 'admin@jbh.cloud',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jbh-cloud/spotify_sync',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import os
import shutil
import sys
import json
import psutil
import jsonschema
from pathlib import Path
from flatdict import FlatDict
from datetime import datetime
from tabulate import tabulate
from typing import List, Union
from jsonschema import validate
from pkg_resources import resource_stream

# local imports
from spotify_sync.appdirs_ import CONFIG
from spotify_sync.common import dump_json, load_json
from spotify_sync.dataclasses import ConfigProfile, Config
from spotify_sync.errors import JsonSchemaInvalid
from spotify_sync.io_ import PersistentDataService

current_schema = json.load(resource_stream("spotify_sync", "data/config_schema.json"))
legacy_schema = json.load(
    resource_stream("spotify_sync", "data/config_schema_legacy.json")
)
config_example = json.load(resource_stream("spotify_sync", "data/config.json.example"))

cfg = None


class ConfigCache:
    def __init__(self):
        self.root: Path = Path(CONFIG)
        self.cache_manifest = self.root / ".profiles.json"
        self.profiles: List[ConfigProfile] = []
        self._init_cache_manifest()
        self._discover_cached_profiles()

    def get(self, name: str) -> Union[ConfigProfile, None]:
        self._discover_cached_profiles()
        return self._get_profile(name)

    def get_path(self, name: str) -> Union[Path, None]:
        for p in self.profiles:
            if p.name == name:
                return p.path

        return None

    def add(self, name: str, path: str, force=False) -> None:
        if " " in name:
            print(f"Profile must not contain spaces!")
            sys.exit(1)

        lookup_profile = self._get_profile(name)
        if lookup_profile is not None and not force:
            print(f'Profile "{name}" already exists!')
            sys.exit(0)

        c_loader = ConfigLoader(config_file=Path(path))
        if not c_loader.is_valid:
            print(f"Config validation error: {c_loader.validation_msg}")
            sys.exit(1)

        # copy config to correct location
        dest = self.root / f"{name}.json"
        shutil.copyfile(path, dest)

        profile = ConfigProfile(name=name, path=dest)
        self.profiles.append(profile)
        self._persist_cache_manifest()

        print(f'Successfully added "{name}" to the config cache')

    def remove(self, name: str) -> None:
        profile = self._get_profile(name)
        if profile is None:
            print(f"{name} does not exist!")
            return

        profile.path.unlink()
        self.profiles.remove(profile)
        self._persist_cache_manifest()
        print(f'Successfully removed "{name}" from the config cache')

    def list(self) -> None:
        summary_data = []
        for p in self.profiles:
            c_loader = ConfigLoader(profile=p)
            conf = c_loader.load()

            summary_data.append(
                [
                    p.name,
                    conf["SPOTIFY_USERNAME"],
                    conf["SPOTIFY_PLAYLISTS_ENABLED"],
                    conf["DEEMIX_DOWNLOAD_PATH"],
                    conf["DEEMIX_MAX_BITRATE"],
                    conf["DEEMIX_SKIP_LOW_QUALITY"],
                    conf["DEEMIX_STRICT_MATCHING"],
                ]
            )

        headers = [
            "profile",
            "spotify_user",
            "playlists",
            "download_path",
            "max_bitrate",
            "skip_low_quality",
            "strict_matching",
        ]

        print()
        print(tabulate(summary_data, headers))
        print()

    def list_paths(self) -> None:
        headers = ["profile", "last_modified", "config", "persistent_data"]
        summary_data = []
        for p in self.profiles:
            c_loader = ConfigLoader(profile=p)
            conf = c_loader.load()
            pd_root = PersistentDataService(Config(p.name, p.path, conf)).root
            summary_data.append(
                [
                    p.name,
                    datetime.fromtimestamp(p.path.lstat().st_mtime).strftime(
                        "%Y-%m-%dT%H%M%S"
                    ),
                    p.path,
                    pd_root,
                ]
            )

        print()
        print(tabulate(summary_data, headers))
        print()

    def generate(self) -> None:
        config_fn = self._get_non_existent_config_fn()
        dump_json(config_fn, config_example, sort_keys=False)
        print(f"Generated example config file: {config_fn}")

    def _discover_cached_profiles(self) -> None:
        self.profiles = []

        manifest_profiles = load_json(self.cache_manifest)
        for name, path in manifest_profiles.items():
            self.profiles.append(ConfigProfile(name, Path(path)))

    def _add_profile(self, name: str, path: Path) -> None:
        source = load_json(path)
        profile = self.root / f"{name}.json"
        dump_json(profile, source)
        self._persist_cache_manifest()

    def _get_profile(self, name) -> Union[ConfigProfile, None]:
        for p in self.profiles:
            if p.name != name:
                continue
            return p
        return None

    def _get_config(self, profile_name: str) -> Union[dict, None]:
        profile = self._get_profile(profile_name)
        if profile is None:
            return None

        return load_json(profile.path)

    @staticmethod
    def _get_non_existent_config_fn() -> Path:
        w_dir = Path(os.getcwd())
        config_file = Path(w_dir / "config.json")

        while config_file.exists():
            parts = config_file.name.split(".")
            if len(parts) == 3:
                i = int(parts[1]) + 1
                config_file = Path(w_dir / (parts[0] + f".{str(i)}." + parts[2]))
            else:
                config_file = Path(w_dir / (parts[0] + ".1." + parts[1]))

        return config_file

    def _init_cache_manifest(self) -> None:
        if not self.cache_manifest.exists():
            dump_json(self.cache_manifest, {}, hidden=True)

    def _persist_cache_manifest(self) -> None:
        manifest = {}
        for p in self.profiles:
            manifest[p.name] = str(p.path)

        dump_json(self.cache_manifest, manifest, hidden=True)


class ConfigLoader:
    def __init__(
        self, profile: ConfigProfile = None, config_file: Path = None, legacy_mode=False
    ):
        self.profile = profile
        self.config_file = config_file
        self.legacy_mode = legacy_mode
        self.is_valid = False
        self.validation_msg = ""
        self._config = {}
        self._load_config()

    def load(self, flatten_settings=True) -> dict:
        if flatten_settings:
            return self._flatten_settings(self._config)
        else:
            return self._config

    def migrate_from_legacy_schema(self) -> dict:
        if not self.is_valid:
            raise JsonSchemaInvalid

        config = config_example

        config["threads"] = self._config["threads"]
        config["deemix"]["arl"] = self._config["deemix"]["arl"]
        config["deemix"]["download_path"] = self._config["deemix"]["download_path"]
        config["deemix"]["max_bitrate"] = self._config["deemix"]["max_bitrate"]
        config["deemix"]["skip_low_quality"] = self._config["deemix"][
            "skip_low_quality"
        ]
        config["spotify"]["client_id"] = self._config["spotify"]["client_id"]
        config["spotify"]["client_secret"] = self._config["spotify"]["client_secret"]
        config["spotify"]["username"] = self._config["spotify"]["username"]
        config["spotify"]["scope"] = self._config["spotify"]["scope"]
        config["spotify"]["redirect_uri_port"] = self._config["spotify"][
            "redirect_uri_port"
        ]
        config["spotify"]["playlists"]["enabled"] = self._config["spotify"][
            "playlists"
        ]["enabled"]
        config["spotify"]["playlists"]["excluded"] = self._config["spotify"][
            "playlists"
        ]["excluded"]
        config["pushover"]["enabled"] = self._config["pushover"]["enabled"]
        config["pushover"]["user_key"] = self._config["pushover"]["user_key"]
        config["pushover"]["api_token"] = self._config["pushover"]["api_token"]
        config["autoscan"]["enabled"] = self._config["autoscan"]["enabled"]
        config["autoscan"]["endpoint"] = self._config["autoscan"]["endpoint"]
        config["autoscan"]["auth_enabled"] = self._config["autoscan"]["auth_enabled"]
        config["autoscan"]["username"] = self._config["autoscan"]["username"]
        config["autoscan"]["password"] = self._config["autoscan"]["password"]

        self.legacy_mode = False
        self._validate(config)

        if self.is_valid:
            return config
        else:
            print(f"Migrated config did not meet the new schema.")
            print(f"Validation error: {self.validation_msg}")
            print(f"Config dump -> {json.dumps(config, indent=True)}")
            sys.exit(1)

    def _load_config(self) -> None:
        if self.profile is None and self.config_file is None:
            raise Exception("Must either specify a config file or config profile!")

        if self.profile is not None:
            try:
                with open(self.profile.path, mode="r", encoding="utf-8") as f:
                    self._config = json.load(f)
            except Exception as ex:
                raise Exception(ex)

            self._validate(self._config)
            if not self.is_valid:
                print(f"Config for profile {self.profile.name} is not valid")
                print(f"Validation error: {self.validation_msg}")
                sys.exit(1)

            self._config["THREADS"] = self._get_threads()
            return

        if self.config_file is not None:
            try:
                with open(self.config_file, mode="r", encoding="utf-8") as f:
                    self._config = json.load(f)
            except Exception as ex:
                raise Exception(ex)

            self._validate(self._config)
            if not self.is_valid:
                print(f"Specified config is not valid")
                print(f"Validation error: {self.validation_msg}")
                sys.exit(1)

            self._get_threads()
            return

        raise Exception(
            "Both profile and config_file are both None, this should never happen"
        )

    def _get_threads(self) -> None:
        # Ensure key case correct
        self._config = {k.lower(): v for k, v in self._config.items()}

        machine_threads = psutil.cpu_count()
        if self._config["threads"] == "":
            return machine_threads

        try:
            threads = int(self._config["threads"])
            threads = (
                threads
                if int(self._config["threads"]) <= machine_threads
                else machine_threads
            )
        except:
            # Failed parsing int from self._config["THREADS"], using max available
            threads = machine_threads

        self._config["threads"] = threads

    def _validate(self, config):
        schema = current_schema if not self.legacy_mode else legacy_schema
        try:
            validate(instance=config, schema=schema)
            self.is_valid = True
        except jsonschema.exceptions.ValidationError as err:
            self.validation_msg = err.message

    @staticmethod
    def _flatten_settings(data: dict):
        return {k.upper(): v for k, v in dict(FlatDict(data, delimiter="_")).items()}


def load() -> Config:
    global cfg
    if cfg is None:
        if os.environ.get("MANUAL_CONFIG_FILE") is not None:
            cl = ConfigLoader(config_file=Path(os.environ["MANUAL_CONFIG_FILE"]))
            cfg = Config(path=cl.config_file, data=cl.load(), profile=None)
        elif os.environ.get("CONFIG_PROFILE") is not None:
            cc = ConfigCache()
            profile = cc.get(os.environ["CONFIG_PROFILE"])

            if profile is None:
                print(f'No profile found with name: {os.environ["CONFIG_PROFILE"]}')
                sys.exit(1)

            cfg = Config(
                path=profile.path,
                data=ConfigLoader(profile=profile).load(),
                profile=profile.name,
            )
            os.environ["LOG_PATH"] = str(PersistentDataService(cfg).get_log_path())
        else:
            raise Exception(
                "Must load config with either a config file or config profile"
            )

    return cfg


def set_env(input_value: str, env_name: str) -> None:
    if input_value is not None:
        os.environ[env_name] = input_value



from click import argument, option, Option, UsageError, Path, STRING, Choice, BOOL
from click_option_group import optgroup, RequiredMutuallyExclusiveOptionGroup

# local imports
from spotify_sync.config import set_env


def notify_config(ctx, param, value):
    if param.name == "config":
        set_env(value, "MANUAL_CONFIG_FILE")

    if param.name == "profile":
        set_env(value, "CONFIG_PROFILE")

    if param.name == "verbose" and value is True:
        set_env("DEBUG", "LOG_LEVEL")


config_option_group = optgroup.group(
    "Input config object",
    cls=RequiredMutuallyExclusiveOptionGroup,
    help="The sources of the config input",
)
config_option_file = optgroup.option(
    "--config",
    type=Path(exists=True),
    callback=notify_config,
    help="Path to a separate config file",
)
config_option_profile = optgroup.option(
    "--profile",
    type=STRING,
    callback=notify_config,
    help="Path to a separate config file",
)

backup_output_dir = option(
    "--out-dir",
    "-o",
    required=False,
    type=Path(exists=True, dir_okay=True, file_okay=False),
    help="Output folder to backup to",
)

restore_zip_file = argument("zip_file", type=Path(exists=True), required=True)

restore_zip_file_profile = option(
    "--new-profile", type=STRING, required=False, help="New profile name to restore to"
)

restore_zip_file_force_overwrite = option(
    "-f",
    "--force",
    default=False,
    is_flag=True,
    required=True,
    help="Overwrites destination profile or persistent files if they exist",
)

spotify_songs_path = option(
    "--spotify",
    required=True,
    type=Path(exists=True),
    help="Path to a Spotify song json file to migrate",
)

processed_songs_path = option(
    "--processed",
    required=True,
    type=Path(exists=True),
    help="Path to a processed song json file to migrate",
)

migrated_profile_name = option(
    "--profile",
    required=True,
    type=STRING,
    help="Name of a (non existent) profile to migrate settings to",
)

migrated_config_schema_mode = option(
    "--config-schema-mode",
    type=Choice(["legacy", "latest"], case_sensitive=False),
    show_default=True,
    default="legacy",
    help="Specify schema of config file",
)

config_list_detailed = option(
    "--detailed",
    "-d",
    required=False,
    is_flag=True,
    help="Display additional information",
)

log_level = option(
    "--verbose",
    "-v",
    required=False,
    is_flag=True,
    callback=notify_config,
    help="Sets verbosity to DEBUG",
)



# defer imports due to click cli speed implications

import sys

from spotify_sync.common import get_temp_file, get_temp_fn, dump_json


class SpotifySyncApp:
    def __init__(self):
        from spotify_sync.config import ConfigCache

        self.config = None
        self.cc = ConfigCache()
        self._logger = None

    def authorize_spotify(self):
        from spotify_sync.io_ import PersistentDataService
        from spotify_sync.spotify_ import SpotifyService

        self._setup(logger=False)

        pd_svc = PersistentDataService(self.config)
        spotify_svc = SpotifyService(self.config, pd_svc)

        spotify_svc.cache_spotify_auth()
        print(f"Successfully cached Spotify OAuth token!")

    def auto(self):
        from spotify_sync.io_ import PersistentDataService
        from spotify_sync.spotify_ import SpotifyService
        from spotify_sync.match import MatchService
        from spotify_sync.autoscan_ import AutoScanService
        from spotify_sync.download import DownloadService
        from spotify_sync.pushover_ import PushoverClient

        self._setup()

        self._logger.info("Script started with auto command")

        pd_svc = PersistentDataService(self.config)
        pushover = PushoverClient(self.config)
        spotify_svc = SpotifyService(self.config, pd_svc)
        match_svc = MatchService(self.config, pd_svc)
        download_svc = DownloadService(self.config, pd_svc, pushover)
        autoscan_svc = AutoScanService(self.config)

        # Checks
        download_svc.preflight()

        pushover.send_message(f"spotify_sync", "Script started with auto command")

        spotify_svc.sync()
        match_svc.process_spotify()
        download_svc.download_missing_tracks()
        autoscan_svc.scan(download_svc.downloaded_song_paths)

        if len(download_svc.downloaded_song_paths) >= 1:
            pushover.send_message(
                "spotify_sync",
                f"Successfully downloaded {len(download_svc.downloaded_song_paths)} new songs(s)",
            )
        pushover.send_message("spotify_sync", f"Script finished")
        self._logger.info("Script finished")
        sys.exit(0)

    def sync_spotify(self):
        from spotify_sync.io_ import PersistentDataService
        from spotify_sync.spotify_ import SpotifyService

        self._setup()

        self._logger.info("Script started with sync-spotify command")

        pd_svc = PersistentDataService(self.config)
        spotify_svc = SpotifyService(self.config, pd_svc)

        spotify_svc.sync()

        self._logger.info("Script finished")
        sys.exit(0)

    def match_spotify(self):
        from spotify_sync.io_ import PersistentDataService
        from spotify_sync.match import MatchService

        self._setup()

        self._logger.info("Script started with match-spotify command")

        pd_svc = PersistentDataService(self.config)
        match_svc = MatchService(self.config, pd_svc)

        match_svc.process_spotify()

        self._logger.info("Script finished")
        sys.exit(0)

    def download_missing(self):
        from spotify_sync.io_ import PersistentDataService
        from spotify_sync.download import DownloadService
        from spotify_sync.pushover_ import PushoverClient

        self._setup()

        self._logger.info("Script started with download-missing command")

        pd_svc = PersistentDataService(self.config)
        pushover = PushoverClient(self.config)
        download_svc = DownloadService(self.config, pd_svc, pushover)

        # Checks
        download_svc.preflight()

        download_svc.download_missing_tracks()
        self._logger.info("Script finished")
        sys.exit(0)

    def scan(self, paths):
        from spotify_sync.autoscan_ import AutoScanService

        self._setup()

        self._logger.info("Script started with manual-scan command")

        autoscan_svc = AutoScanService(self.config)
        autoscan_svc.scan(paths)

        self._logger.info("Script finished")

    def playlist_stats(self):
        from spotify_sync.io_ import PersistentDataService
        from spotify_sync.spotify_ import SpotifyService

        self._setup(logger=False)

        pd_svc = PersistentDataService(self.config)
        spotify_svc = SpotifyService(self.config, pd_svc)

        spotify_svc.display_playlist_stats()

    def validate_downloaded_files(self):
        pass
        # from spotify_sync.io_
        #
        # self._logger.info('Script started with validate-downloaded-files command')
        # missing = io_.get_missing_files()
        # self._logger.info(f'Found {str(missing)} missing songs')
        # self._logger.info('Script finished')

    def failed_download_stats(self):
        from spotify_sync.io_ import PersistentDataService
        from spotify_sync.download import DownloadService
        from spotify_sync.pushover_ import PushoverClient

        self._setup(logger=False)

        pd_svc = PersistentDataService(self.config)
        DownloadService(
            self.config, pd_svc, PushoverClient()
        ).display_failed_download_stats()

    def cache_config_profile(self, name, path):
        self.cc.add(name, path)

    def remove_config_profile(self, name):
        self.cc.remove(name)

    def list_config_profiles(self):
        self.cc.list()

    def list_config_profile_paths(self):
        self.cc.list_paths()

    def generate_example_config(self):
        self.cc.generate()

    @staticmethod
    def migrate_config():
        import json
        import os
        from pathlib import Path

        from spotify_sync.config import ConfigLoader

        if os.environ.get("MANUAL_CONFIG_FILE") is None:
            print(f"Must provide a config file to migrate")
            sys.exit(1)

        cl = ConfigLoader(
            config_file=Path(os.environ["MANUAL_CONFIG_FILE"]), legacy_mode=True
        )
        new_config = cl.migrate_from_legacy_schema()

        output = Path(os.getcwd()) / get_temp_fn(extension=".json")
        dump_json(output, new_config)
        print(f"Successfully migrated config -> {output}")

    @staticmethod
    def migrate_to_profile(spotify, processed, profile_name, config_is_legacy):
        import shutil
        import json
        import os
        from pathlib import Path

        from spotify_sync.dataclasses import Config
        from spotify_sync.io_ import PersistentDataService
        from spotify_sync.config import ConfigLoader, ConfigCache
        from spotify_sync.common import get_temp_file

        if os.environ.get("MANUAL_CONFIG_FILE") is None:
            print(f"Must provide a config file to migrate")
            sys.exit(1)

        cfg_file = Path(os.environ["MANUAL_CONFIG_FILE"])

        cc = ConfigCache()
        cl = ConfigLoader(config_file=cfg_file, legacy_mode=config_is_legacy)

        if config_is_legacy:
            config = cl.migrate_from_legacy_schema()
            cfg_file = get_temp_file()
            with open(cfg_file, mode="w", encoding="utf-8") as f:
                json.dump(config, f)

        cc.add(profile_name, cfg_file)
        profile = cc.get(profile_name)

        pd_svc = PersistentDataService(
            Config(path=profile.path, data=cl.load(), profile=profile.name)
        )

        # TODO: handle path already exist
        shutil.copy(spotify, pd_svc.get_spotify_songs_path())
        shutil.copy(processed, pd_svc.get_processed_songs_path())

        if config_is_legacy:
            cfg_file.unlink()

        print(f'Migrated files successfully to profile "{profile_name}"')

    def local_backup(self, out_dir: str):
        from spotify_sync.errors import SnapshotFileNotExists
        from spotify_sync.io_ import PersistentDataService
        from spotify_sync.backup.providers import FileSystemBackupProvider

        self._setup(logger=False)

        try:
            snapshot = PersistentDataService(self.config).gather_snapshot_files()
            provider = FileSystemBackupProvider()
            provider.backup(out_dir, snapshot)
        except SnapshotFileNotExists:
            print(
                f"Failed local backup - One or more required snapshot files do not exist!"
            )
            sys.exit(1)
        except:
            raise

    @staticmethod
    def local_restore(zip_file: str, force: bool, new_profile: str):
        from pathlib import Path
        from spotify_sync.errors import RestoreZipNotExists
        from spotify_sync.backup.providers import FileSystemBackupProvider

        try:
            provider = FileSystemBackupProvider()
            provider.restore(
                zip_file=Path(zip_file), force=force, new_profile=new_profile
            )
        except RestoreZipNotExists:
            print(f'Specified zip to restore does not exist: "{zip_file}"')
            sys.exit(1)
        except:
            raise

    def _notify_user_config(self):
        self._logger.info(
            f"Loading config from profile: {self.config.profile}"
        ) if self.config.profile is not None else self._logger.info(
            f"Loading config from file {self.config.path}"
        )

    def _load_config(self):
        from spotify_sync.config import load as load_config

        self.config = load_config()

    def _setup_logger(self):
        from spotify_sync.log import get_logger

        self._logger = get_logger().getChild("SpotifySync")

    def _setup(self, logger=True):
        self._load_config()
        if logger:
            self._setup_logger()
            self._notify_user_config()



import json
from pathlib import Path
from typing import Dict

# local imports
from spotify_sync.appdirs_ import DATA
from spotify_sync.config import Config
from spotify_sync.dataclasses import SpotifySong, ProcessedSong, BackupSnapshot
from spotify_sync.common import load_json, dump_json, as_spotify_song, as_processed_song
from spotify_sync.errors import SnapshotFileNotExists


class PersistentDataService:
    def __init__(self, app_config: Config):
        self.config = app_config
        self.root: Path = self._ensure_root_dir(app_config)

    def verify_files(self):
        files = [self.get_spotify_songs_path(), self.get_processed_songs_path()]

        for f in files:
            if not Path(f).is_file():
                with open(f, mode="w", encoding="utf-8") as fp:
                    json.dump({}, fp)

    def load_spotify_songs(self) -> Dict[str, SpotifySong]:
        file = self.get_spotify_songs_path()
        return as_spotify_song(load_json(file))

    def load_processed_songs(self) -> Dict[str, ProcessedSong]:
        file = self.get_processed_songs_path()
        return as_processed_song(load_json(file))

    def persist_spotify_songs(self, songs) -> None:
        file = self.get_spotify_songs_path()
        file.parent.mkdir(exist_ok=True)
        dump_json(file, songs)

    def persist_processed_songs(self, songs) -> None:
        file = self.get_processed_songs_path()
        file.parent.mkdir(exist_ok=True)
        dump_json(file, songs)

    def get_spotify_oauth(self):
        return self.get_spotify_oauth_path()

    def gather_snapshot_files(self) -> BackupSnapshot:
        snapshot = BackupSnapshot(
            profile=self.config.profile,
            username=self.config.data["SPOTIFY_USERNAME"],
            config=self.config.path,
            spotify=self.get_spotify_songs_path(),
            processed=self.get_processed_songs_path(),
            playlists=self.get_playlist_mapping_path(),
            oauth=self.get_spotify_oauth_path(),
        )

        # TODO: Allow partial backup and restore
        kill = False
        if not snapshot.config.exists():
            print(f"Cannot find config file: {snapshot.config}")
            kill = True
        if not snapshot.spotify.exists():
            print(f"Cannot find spotify songs file: {snapshot.spotify}")
            kill = True
        if not snapshot.processed.exists():
            print(f"Cannot find processed songs file: {snapshot.processed}")
            kill = True
        if not snapshot.playlists.exists():
            print(f"Cannot find playlists file: {snapshot.playlists}")
            kill = True
        if not snapshot.oauth.exists():
            print(f"Cannot find spotify oauth cache file: {snapshot.oauth}")
            kill = True

        if kill:
            raise SnapshotFileNotExists

        return snapshot

    def get_spotify_songs_path(self) -> Path:
        return self.root / f'spotify-{self.config.data["SPOTIFY_USERNAME"]}.json'

    def get_processed_songs_path(self) -> Path:
        return self.root / f'processed-{self.config.data["SPOTIFY_USERNAME"]}.json'

    def get_playlist_mapping_path(self) -> Path:
        return (
            self.root / f'playlist-mapping-{self.config.data["SPOTIFY_USERNAME"]}.json'
        )

    def get_spotify_oauth_path(self) -> Path:
        return (
            self.root / f'.spotify-oauth-cache-{self.config.data["SPOTIFY_USERNAME"]}'
        )

    def get_log_path(self) -> Path:
        return self.root / "logs" / "spotify_sync.log"

    def persist_playlist_mapping(self, mapping):
        file = Path(self.get_playlist_mapping_path())
        dump_json(file, mapping)

    @staticmethod
    def _ensure_root_dir(config: Config) -> Path:
        if config.profile is not None:
            root = Path(DATA) / (config.data["SPOTIFY_USERNAME"] + f"_{config.profile}")
        else:
            root = Path(DATA) / config.data["SPOTIFY_USERNAME"]

        root.mkdir(parents=True, exist_ok=True)
        return root



from pathlib import Path
from appdirs import AppDirs

_dirs = AppDirs("spotify_sync", "jbh-cloud")

CONFIG = Path(_dirs.user_config_dir)
DATA = Path(_dirs.user_data_dir)

Path(CONFIG).mkdir(parents=True, exist_ok=True)
Path(DATA).mkdir(parents=True, exist_ok=True)



import os
import sys
import hmac
import json
import shutil
import base64
import hashlib
from pathlib import Path
from datetime import datetime
from typing import Union
from zipfile import ZipFile, ZIP_DEFLATED

# local imports
from spotify_sync.common import get_temp_dir
from spotify_sync.config import ConfigLoader, ConfigCache
from spotify_sync.dataclasses import BackupSnapshot, BackupManifest, Config
from spotify_sync.errors import RestoreZipNotExists
from spotify_sync.io_ import PersistentDataService

INTEGRITY_BLOB = "0563debb-4660-4696-b968-e282c68c568b"


class BackupProvider:
    def __init__(self, **kwargs):
        self.snapshot: Union[BackupSnapshot, None] = None

    def backup(self, **kwargs):
        pass

    def restore(self, **kwargs):
        pass

    def restore_zip_file(
        self,
        zip_file: Path,
        new_profile: Union[str, None] = None,
        overwrite_files=False,
    ):
        if not zip_file.exists() or not zip_file.is_file():
            raise RestoreZipNotExists

        try:
            temp = get_temp_dir()
            extract_dest = temp / zip_file.name
            extract_dest.mkdir(exist_ok=False)
            with ZipFile(zip_file, "r") as zipObj:
                zipObj.extractall(extract_dest)

            manifest = self._parse_manifest(extract_dest / ".manifest.json")
            if self._manifest_tampered(manifest):
                print(f"Backup has been manually tampered with, exiting..")
                sys.exit(1)

            snapshot = BackupSnapshot(
                manifest.profile, manifest.spotify_user, None, None, None, None, None
            )

            if new_profile is not None:
                snapshot.profile = new_profile

            for f in extract_dest.glob("*.json"):
                if f.name == "config.json":
                    snapshot.config = f
                if f.name == "spotify.json":
                    snapshot.spotify = f
                if f.name == "processed.json":
                    snapshot.processed = f
                if f.name == "playlists.json":
                    snapshot.playlists = f
                if f.name == ".spotify-oauth.json":
                    snapshot.oauth = f

            if (
                snapshot.config is None
                or snapshot.spotify is None
                or snapshot.processed is None
            ):
                print(f"Mandatory files are not present in backup, unable to restore..")
                sys.exit(1)

            self._restore_snapshot(snapshot, force=overwrite_files)
            print(
                f'Successfully restored persistent data for Spotify user "{snapshot.username}" from backup'
            )

        except Exception as ex:
            print(f"Failed restore, reason: {ex}")
            sys.exit(1)

        finally:
            try:
                shutil.rmtree(extract_dest)
            except UnboundLocalError:
                pass  # folder not created yet
            except Exception as ex:
                print(
                    f"Failed to remove temp dir: {extract_dest}. Please clean this up manually."
                )

    @staticmethod
    def _restore_snapshot(snapshot, force=False) -> None:
        cc = ConfigCache()
        c_loader = ConfigLoader(config_file=Path(snapshot.config))
        config = Config(profile=None, path=snapshot.config, data=c_loader.load())

        if snapshot.profile is not None:
            config.profile = snapshot.profile
            cc.add(snapshot.profile, snapshot.config, force=force)

        pd_svc = PersistentDataService(config)
        file_map = {}
        if snapshot.spotify is not None:
            file_map[snapshot.spotify] = pd_svc.get_spotify_songs_path()
        if snapshot.processed is not None:
            file_map[snapshot.processed] = pd_svc.get_processed_songs_path()
        if snapshot.playlists is not None:
            file_map[snapshot.playlists] = pd_svc.get_playlist_mapping_path()
        if snapshot.oauth is not None:
            file_map[snapshot.oauth] = pd_svc.get_spotify_oauth_path()

        for src, dest in file_map.items():
            if not dest.exists() or force:
                shutil.copy(src, dest)  # For Python 3.8+.
            else:
                print(
                    f"Skipping restore of {dest}, file exists and --force is not specified"
                )

    @staticmethod
    def _parse_manifest(file: Path) -> BackupManifest:
        if not file.exists() or not file.is_file():
            print(
                f"Unable to find manifest in backup, are you use this is a spotify_sync backup?"
            )
            sys.exit(1)

        with open(file, mode="r", encoding="utf-8") as f:
            manifest = json.load(f, object_hook=lambda d: BackupManifest(**d))

        return manifest

    @staticmethod
    def _manifest_tampered(manifest: BackupManifest) -> bool:
        message = (manifest.spotify_user + manifest.timestamp).encode("utf-8")
        dig = hmac.new(
            INTEGRITY_BLOB.encode("utf-8"), msg=message, digestmod=hashlib.sha256
        ).digest()
        should_match = base64.b64encode(dig).decode()
        return not manifest.checksum == should_match

    def add_files_to_zip(self, output_zip: Path) -> None:
        manifest = self._generate_manifest()

        with ZipFile(
            output_zip, "w", compression=ZIP_DEFLATED, compresslevel=9
        ) as zipObj:
            zipObj.write(self.snapshot.config, arcname="config.json")
            zipObj.write(self.snapshot.spotify, arcname="spotify.json")
            zipObj.write(self.snapshot.processed, arcname="processed.json")
            zipObj.write(self.snapshot.playlists, arcname="playlists.json")
            zipObj.write(self.snapshot.oauth, arcname=".spotify-oauth.json")
            zipObj.writestr(
                ".manifest.json",
                json.dumps(
                    manifest, indent=4, sort_keys=True, default=lambda obj: obj.__dict__
                ),
            )

    def _generate_manifest(self) -> BackupManifest:
        with open(self.snapshot.config, mode="r", encoding="utf-8") as f:
            config = json.load(f)

        timestamp = datetime.utcnow().strftime("%Y-%m-%dT%H%M%S")
        username = config["spotify"]["username"]
        checksum = self._get_manifest_checksum(timestamp, username)

        return BackupManifest(
            profile=self.snapshot.profile,
            spotify_user=username,
            timestamp=timestamp,
            checksum=checksum,
        )

    @staticmethod
    def _get_manifest_checksum(timestamp, username):
        message = (username + timestamp).encode("utf-8")
        dig = hmac.new(
            INTEGRITY_BLOB.encode("utf-8"), msg=message, digestmod=hashlib.sha256
        ).digest()
        return base64.b64encode(dig).decode()  # py3k-mode


class FileSystemBackupProvider(BackupProvider):
    def __init__(self):
        super().__init__()

    def backup(self, out_dir: Union[str, None], snapshot: BackupSnapshot):
        self.snapshot = snapshot
        try:
            export_fn = (
                Path(os.getcwd())
                if out_dir is None
                else Path(out_dir)
                / (
                    "spotify_sync_backup_"
                    + datetime.utcnow().strftime("%Y-%m-%dT%H%M%S")
                    + ".zip"
                )
            )
            self.add_files_to_zip(export_fn)
            print(f"Successfully backed up. {export_fn}")
            sys.exit(0)
        except Exception as ex:
            print(f"Failed creating backup, error was: {ex}")
            sys.exit(1)

    def restore(self, zip_file: Path, force=False, new_profile=Union[str, None]):
        self.restore_zip_file(zip_file, overwrite_files=force, new_profile=new_profile)
